/* Gestion des signaux */

#include <signal.h>
#include "mlvalues.h"
#include "signals.h"
#include "globals.h"
#include "alloc.h"
#include "memory.h"
#include "fail.h"

int signal_is_pending;
int block_signals;
int raise_on_signal;
value pending_signal;

#ifndef NSIG
#define NSIG 32
#endif

void init_signals()
{
  value signal_table;
  int i;

  signal_is_pending = 0;
  pending_signal = MLINT (0);
  block_signals = 0;
  raise_on_signal = 0;
  SMALL_ALLOC(signal_table, MAKE_HEADER(NSIG, 0), NSIG);
  for (i = 0; i < NSIG; i++)
    FIELD(signal_table, i) = atom(BOTTOM_EXN);
  MODIFY(&FIELD(global_data, SIGNAL_TABLE), signal_table);
}

void raise_pending_signal()
{
  value exc;
  exc = pending_signal;
  pending_signal = MLINT(1);
  signal_is_pending = 0;
  raise_on_signal = 0;
  mlraise(exc);
}

#ifdef ANSI
static void signal_handler(sig)
#else
static signal_handler(sig)
#endif
     int sig;
{
  value signal_table;

#ifndef BSD_SIGNALS
  (void) signal (sig, signal_handler);
#endif
  if (block_signals) return;
  signal_table = FIELD(global_data, SIGNAL_TABLE);
  if (sig >= 0 && sig < SIZE(signal_table)) {
    pending_signal = FIELD(signal_table, sig);
    signal_is_pending = 1;
    if (raise_on_signal)
      raise_pending_signal();
  }
}

value set_signal(sign, behavior)	/* ML */
     value sign, behavior;
{
  value signal_table;
  int sig;

  signal_table = FIELD(global_data, SIGNAL_TABLE);
  sig = CINT(sign);
  if (sig < 0) {
    block_signals = (TAG(behavior) == 1);
  } else {
    switch(TAG(behavior)) {
    case 0:                       /* Signal_default */
      signal(sig, SIG_DFL);
      FIELD(signal_table, sig) = atom(BOTTOM_EXN);
      break;
    case 1:                       /* Signal_ignore */
      signal(sig, SIG_IGN);
      FIELD(signal_table, sig) = atom(BOTTOM_EXN);
      break;
    case 2:                       /* Signal_raise */
      MODIFY(&FIELD(signal_table, sig), FIELD(behavior, 0));
      signal(sig, signal_handler);
      break;
    }
  }
  return atom(0);
}

